/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev$ $LastChangedDate$
 *
 * Genlex - generic lexer.
 */

open Genlex;
open Unitype;

struct op_node {
    <char, op_node>Ptrhash::t next;
    string leaf;
}

struct mode {
    *(bool (char)) is_number_beg;
    *(bool (char)) is_number_body;
    *(bool (char)) is_id;
    *(bool (char)) is_op;
    *(bool (char)) is_blank;
    
    <char, op_node>Ptrhash::t operators;
}

mode make_mode()
{
    return {
	is_id		= fun (c) (is_ascii_alnum(c) || 
				   Hacks::same_chars(c, '_')),
	is_number_beg	= is_ascii_number,
	is_number_body  = fun (c) (is_ascii_xnumber(c) || 
			           String::contains("x.eE+-", c)),
	is_op		= fun (c) (is_ascii(c) && 
				   !Hacks::same_chars(c, '_') &&
				   (is_punct(c) || is_symbol(c))),
	is_blank	= fun (c) (is_ascii_blank(c) || 
				   Hacks::same_chars(c, '\n')),
	
	operators	= Ptrhash::create(0)
    };
}

void set_is_number_beg(mode m, *(bool (char)) f) { m.is_number_beg = f; }
void set_is_number_body(mode m, *(bool (char)) f) { m.is_number_body = f; }
void set_is_id(mode m, *(bool (char)) f) { m.is_id = f; }
void set_is_op(mode m, *(bool (char)) f) { m.is_op = f; }
void set_is_blank(mode m, *(bool (char)) f) { m.is_blank = f; }

void add_operator(mode m, string op)
{
    _ l = Array::to_list(String::to_char_array(op));
    
    def add_rec(n, lst)
    {
	switch lst {
	case Cons(hd, tl):
	    if (!Ptrhash::mem(n.next, hd))
		Ptrhash::add(n.next, hd, {next = Ptrhash::create(0), 
				          leaf = ""});
	    add_rec(Ptrhash::find(n.next, hd), tl);
	    
	case Nil:
	    n.leaf = op;
	}
    }
    
    switch l {
    case Cons(c, rest):
	if (!m.is_op(c))
	    raise Failure("Genlex::add_operator");
	switch rest {
	case Nil:
	    skip;
	case _:
	    if (!Ptrhash::mem(m.operators, c))
		Ptrhash::add(m.operators, c, {next = Ptrhash::create(0), 
					      leaf = ""});
	    add_rec(Ptrhash::find(m.operators, c), rest);
	}

    case Nil:
	raise Failure("Genlex::add_operator");
    }
}

struct t {
    <char>Array::t buf;
    int buf_pos;
    <Stdio::FILE>option buf_refill;
    
    int line, column;
}

t of_string(string s)
{
    _ r = {buf = Array::empty(0),
	   buf_pos = 0,
	   buf_refill = None,
	   line = 1,
	   column = 1};
    
    refill(r, s);

    return r;
}

t of_stream(Stdio::FILE f)
{
    return {buf = Array::empty(0),
	    buf_pos = 0,
	    buf_refill = Some(f),
	    line = 1,
	    column = 1};
}

*(int, int) get_location(t l)
{
    return (l.line, l.column);
}

void refill(t l, string s)
{
    l.buf = String::to_char_array(s);
    l.buf_pos = 0;

    bool seen_cr = false;
    bool seen_nl = false;

    def check(c) 
    {
	if (Hacks::same_chars(c, '\n')) 
	    seen_nl = true;
	if (Hacks::same_chars(c, '\r')) 
	    seen_cr = true;
    }
//puts("refill '" + s + "'\n");
    
    Array::iter(check, l.buf);

    if (seen_cr) {
	_ len = Array::length(l.buf);
	
	if (seen_nl) {
	    _ tmp = Array::empty(0);
	    bool did_cr = false;
	    for (_ i = 0; i < len; i++) {
		_ c = Array::get(l.buf, i);
		if (Hacks::same_chars(c, '\n')) {
		    did_cr = false;
		} else if (Hacks::same_chars(c, '\r') && i != len - 1) {
		    did_cr = true;
		    continue;
		} else if (Hacks::same_chars(c, '\r')) {
		    c = '\n';
		} else if (did_cr) {
		    Array::append(tmp, '\n');
		    did_cr = false;
		}

		Array::append(tmp, c);
	    }
	    
	    l.buf = tmp;
	} else {
	    for (_ i = 0; i < len; i++) {
		if (Hacks::same_chars(Array::get(l.buf, i), '\r'))
		    Array::set(l.buf, i, '\n');
	    }
	}
    }
//puts("end refill\n");
}

void maybe_refill(t l)
{
    if (l.buf_pos >= Array::length(l.buf))
	switch l.buf_refill {
	case Some(f):
	    refill(l, Stdio::fgets(f));

	    if (Array::length(l.buf) == 0)
		raise End_of_file;	// what the heck?

	case None:
	    raise End_of_file;
	}
}

char peek_char(t l)
{
    maybe_refill(l);
    return Array::get(l.buf, l.buf_pos);
}

char get_char(t l)
{
    maybe_refill(l);
    
    _ c = Array::get(l.buf, l.buf_pos++);
    if (Hacks::same_chars(c, '\n')) {
	l.line++;
	l.column = 1;
    } else {
	l.column++;
    }

//puts("char `" + String::make(1, c) + "'\n");
    
    return c;
}

void shift_char(t l)
{
    get_char(l);
}

token get_token(t l, mode m)
{
    try {
	_ c = peek_char(l);
	_ buf = Array::empty(100);

	string get_while_true(*(bool (char)) f)
	{
	    Array::append(buf, c);
	    shift_char(l);
	    try {
		while (f(peek_char(l)))
		    Array::append(buf, get_char(l));
	    } with { case End_of_file: }
	    
	    return String::of_char_array(buf);
	}

	if (m.is_number_beg(c)) {
	    return Number(get_while_true(m.is_number_body));
	} else if (m.is_op(c)) {
	    string op = String::make(1, c);
	    int off = 0;

	    shift_char(l);
	    
	    try {
		_ n = Ptrhash::find(m.operators, c);
		
		for (;;) {
		    n = Ptrhash::find(n.next, peek_char(l));
		    if (n.leaf != "") {
			op = n.leaf;
			off = 0;
		    } else {
			off++;
		    }
		    shift_char(l);
		}
	    } with { 
	    case End_of_file: 
	    case Not_found:
//    puts("off="+itoa(off)+","+itoa(l.buf_pos)+"\n");
		l.buf_pos = l.buf_pos - off;
		l.column -= off;
		if (l.buf_pos < 0)
		    l.buf_pos = 0;
		if (l.line < 1)
		    l.line = 1;
		return Operator(op);
	    }
	} else if (m.is_blank(c)) {
	    return Blank(get_while_true(m.is_blank));
	} else if (m.is_id(c)) {
	    return Id(get_while_true(m.is_id));
	} else {
	    shift_char(l);
	    return Junk(c);
	}
    } with {
    case End_of_file:
	return Eof;
    }
}

// vim:sts=4:sw=4
