/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 964 $ $LastChangedDate: 2002-09-14 10:19:57 +0200 (Sat, 14 Sep 2002) $
 *
 * Standard library for Gont, Hashtbl module interface.
 *
 * Mutable hash table.
 */

/** This module provides mutable hash tables from string to any type.  */

/** Type of hash tables holding values of type ['a].  */
type <'a>t;

/** Create new hash, empty table.  [size_hint] is order of initial size of 
    hash table.  Hash table will grow as necessary.  */
<'a>t create(int size_hint);

/** Add element [v] with key [k] to [h].  If there is already element with
    the same key, this new element shadows it, subsequent calls to 
    [find(h, k)] will return [v].  However after [remove(h, k)], [find(h, k)]
    will return shadowed element.  */
void add(<'a>t h, string k, 'a v);

/** Look element with key [k] in [h].  Raise [Not_found] if element is not
    in hash table.  Otherwise return it.  */
'a find(<'a>t h, string k);

/** Remove one element with key [k] from [h].  Raise [Not_found] if element 
    is not in hash table.  */
void remove(<'a>t h, string k);

/** Remove all elements with key [k] from [h].  If no such elements are found,
    do nothing.  */
void remove_all(<'a>t h, string k);

/** Return list of all elements with key [k] from [h].  If no such elements are
    found return [Nil].  */
<'a>list find_all(<'a>t h, string k);

/** Return [true] iff element with key [k] is in [h].  */
bool mem(<'a>t h, string k);

/** Return number of elements stored in [h].  */
int length(<'a>t h);

/** Execute [f] on each element of [h], passing it its key and value.  
    It is undefined what elements will be visited if [h] is modified 
    (elements are inserted or removed) during iteration.  */
void iter(<'a>t h, *(void (string, 'a)) f);

/** Return copy of [h].  Time of this operation is [O(length(h))].  */
<'a>t copy(<'a>t h);

/** Same as [remove_all(h, k); add(h, k, v);].  */
void replace(<'a>t h, string k, 'a v);

/** Return list of keys of elements stored in [h].  Keys are returned in
    arbitrary order, however [keys(h)] and [value(h)] will maintain the same
    order if there are no insertions nor deletions on [h] between them.  */
<string>list keys(<'a>t h);

/** Return list of values of elements stored in [h].  See Hashtbl::keys()
    for ordering notes.  */
<'a>list values(<'a>t);

/** Create new hash table and add each (key, value) pair from [assoc].
    Return created hashtable.  */
<'a>t of_assoc_list(<*(string, 'a)>list assoc);

/** Return prime number [p], such that [p >= x] and [p] is in about the
    same binary order of magnitude as [x].  Additionally [p >= 13]. 
    It is undefined for [x > 2^30].  */
int prime_around(int x) = "Hashtbl__prime_around";
