/*
Syntax recognized:

decl: 
  SYMBOL ( '{' BALANCED '}' )? ( '[' BALANCED ']' )? ';'    // terminal
  SYMBOL '{' BALANCED '}' ':' rule | rule | ... ';'
  SYMBOL : 'list' SYMBOL ('separator' SYMBOL)? ('not_empty')? ';'
  SYMBOL : 'optional' SYMBOL ';'

rule: 
  ((SYMBOL | CHAR)  ( '[' BALANCED ']' )?)* ('%' NOT-% '%')? body

body: 
  '{' BALANCED '}'
  '(' BALANCED ')'
*/

open Stdio;

string module = "Parser";
string prefix = "gyy_";
string start_symbol = "";

string location_type = "";

string g_hd = "";
string g_tl = "";
string gi_hd = "";
string gi_tl = "";
string y_hd = "";
string y_tl = "";
string c_hd = "";
string c_tl = "";

// datatypes
struct rhs_entry {
    string yacc_name;
    string gont_name;
}

struct rule {
    <rhs_entry>list args;
    string annotation;
    string body;
}

struct symbol {
    string name;
    string ty;
    bool terminal;
    string terminal_name;
    <rule>list rules;
    int id;
}

<symbol>Hashtbl::t symbols = Hashtbl::create(133);
<symbol>list symbol_list = [];
<char, void>Ptrhash::t one_char_terminals = Ptrhash::create(133);

// lexer
union token {
    string Id;		// foo or 'f'
    string Name;	// [foo]
    string Return;	// (foo)
    string Body;	// {foo}
    string Annotation;	// %prec foo%
    void Colon;		// :
    void Semicolon;	// ;
    void Or;		// |
}

struct lexer {
    FILE input;
    <char>Array::t line;
    int pos;
}

int lines_read = 0;

lexer make_lexer(FILE f)
{
    return {input = f, line = Array::empty(0), pos = 0};
}

token get_token(lexer l)
{
    bool is_id_char(char c)
    {
	int i = int_of_char(c);
	
	bool in_range(char beg, char end)
	{
	    return (i >= int_of_char(beg) && i <= int_of_char(end));
	}

	return in_range('a', 'z') || in_range('A', 'Z') || 
	       in_range('0', '9') || int_of_char(c) == int_of_char('_');
    }
    

    void next_line()
    {
	string eat_lines(string s)
	{
	    for (;;) {
		string line = fgets(l.input);

		lines_read++;
		
		if (line == "%end\n")
		    break;

		s = s + line;
	    }

	    return s;
	}

	l.line = String::to_char_array(fgets(l.input));
	l.pos = 0;
	lines_read++;

	if (int_of_char(Array::get(l.line, 0)) == int_of_char('%')) {
	    switch String::of_char_array(l.line) {
	    case "%g_hd\n": g_hd = eat_lines(g_hd);
	    case "%g_tl\n": g_tl = eat_lines(g_tl);
	    case "%gi_hd\n": gi_hd = eat_lines(gi_hd);
	    case "%gi_tl\n": gi_tl = eat_lines(gi_tl);
	    case "%y_hd\n": y_hd = eat_lines(y_hd);
	    case "%y_tl\n": y_tl = eat_lines(y_tl);
	    case "%c_hd\n": c_hd = eat_lines(c_hd);
	    case "%c_tl\n": c_tl = eat_lines(c_tl);
	    case s if (String::sub(s, 0, 7) == "%start "):
		start_symbol = String::sub(s, 7, String::length(s) - 8);
		
	    case s if (String::sub(s, 0, 15) == "%location_type "):
		location_type = String::sub(s, 15, String::length(s) - 16);

	    case s:
		die("unknown directive " + s);
	    }

	    next_line();
	}
    }
    
    char getc()
    {
	if (l.pos >= Array::length(l.line))
	    next_line();

	return Array::get(l.line, l.pos++);
    }

    string get_balanced(char beg, char end)
    {
	<char>Array::t line = l.line;
	int pos = l.pos;
	
	<char>Array::t ret = Array::empty(100);

	int balance = 1;
	
	for (;;) {
	    if (pos >= Array::length(line)) {
		lines_read++;
		line = String::to_char_array(fgets(l.input));
		pos = 0;
	    }

	    char c = Array::get(line, pos++);

	    if (int_of_char(c) == int_of_char(end))
		balance--;
	    else if (int_of_char(c) == int_of_char(beg))
		balance++;
	    
	    if (balance == 0)
		break;
		
	    Array::append(ret, c);
	}

	l.line = line;
	l.pos = pos;

	return String::of_char_array(ret);
    }

    for (;;) {
	char c = getc();
	string r = String::make(1, c);

	if (is_id_char(c)) {
	    id: for (;;) {
		c = getc();
		if (!is_id_char(c))
		    break id;
		// hopefully this won't get too large
		r = r + String::make(1, c);
	    }
	    l.pos--;
	    return Id(r);
	} else {
	    // sigh... cannot pattern match chars
	    switch r {
	    case "|":	return Or;
	    case ";":	return Semicolon;
	    case ":":	return Colon;
	    case "'":	return Id("'" + get_balanced('\'', '\'') + "'");
	    case "%":	return Annotation("%" + get_balanced('%', '%'));
	    case "(":	return Return(get_balanced('(', ')'));
	    case "[":	return Name(get_balanced('[', ']'));
	    case "{":	return Body(get_balanced('{', '}'));
	    // skip comments
	    case "#":
	    case "/":	next_line();
	    // and white space
	    case " ":
	    case "\n":
	    case "\r":
	    case "\t":	skip;
	    // choke on other
	    case _:	die("illegal character in input '" + r + "'");
	    }
	}
    }
}


bool read_rule(lexer l, symbol s)
{
    rule r = {args = Nil, body = "", annotation = ""};

    void push()
    {
	r.args = List::rev(r.args);
	s.rules = Cons(r, s.rules);
    }

    entries: for (;;)
	switch get_token(l) {
	case Id("empty"):
	    continue entries;
	    
	case Id(i):
	    r.args = Cons(({yacc_name = i, gont_name = ""}), r.args);
	    if (String::sub(i, 0, 1) == "'") {
		if (String::length(i) != 3 || String::sub(i, 2, 1) != "'")
		    die("character-terminal is not 1-char long");
		Ptrhash::replace(one_char_terminals, String::get(i, 1), ());
	    }
	    
	case Name(i):
	    switch r.args {
	    case Cons(hd, _):
		if (hd.gont_name != "")
		    die("gont name already specified");
		hd.gont_name = i;
	    case Nil:
		die("no yacc name to add gont name");
	    }
	
	case Annotation(a):
	    r.annotation = r.annotation + " " + a;
	    
	case Body(b):
	    r.body = b;
	    break entries;
	
	case Return(b):
	    r.body = "return " + b + ";\n";
	    break entries;

	case Or:
	    push();
	    return true;
	    
        case Semicolon:
	    push();
	    return false;

	case _:
	    die("unexpected token");
	}

    push();
    
    switch get_token(l) {
    case Or:
	return true;
    case Semicolon:
	return false;
    case _:
	return die("expected | or ;");
    }
}

string get_id(lexer l)
{
    switch get_token(l) {
    case Id(i):
	return i;
    case _:
	return die("expected id");
    }
}

void read_symbol(lexer l)
{
    token t;
   
    void read_list(symbol s)
    {
	_ of = get_id(l);
	_ sep = None;
	bool non_empty = false;
	_ r1, r2;

	for (;;) {
	    switch get_token(l) {
	    case Semicolon:		break;
	    case Id("separator"):	sep = Some(get_id(l));
	    case Id("non_empty"):	non_empty = true;
	    case _:			die("bad list syntax");
	    }
	}
	
	try {
	    let o = Hashtbl::find(symbols, of) in
	    s.ty = "<" + o.ty + ">list";
	} with {
	case Not_found:
	    die("no such symbol " + of);
	}
	
	switch sep {
	case None:
	    r1 = ({annotation = "", 
		   body = "return Cons(hd, tl);\n",
		   args = List::cons2(({yacc_name = of, 
					  gont_name = "hd"}),
					({yacc_name = s.name, 
					  gont_name = "tl"}))});
	    if (non_empty)
		r2 = ({annotation = "", 
		       body = "return Cons(hd, Nil);\n",
		       args = List::cons1(({yacc_name = of, 
					  gont_name = "hd"}))});
	    else
		r2 = ({annotation = "", 
		       body = "return Nil;\n",
		       args = Nil});
	    
	case Some(sep_sym):
	    string rec_name = s.name;
	    if (!non_empty) rec_name = "ne_" + s.name;
	    r1 = ({annotation = "", 
		   body = "return Cons(hd, tl);\n",
		   args = List::cons3(({yacc_name = of, 
					gont_name = "hd"}),
				      ({yacc_name = sep_sym, 
					gont_name = ""}),
				      ({yacc_name = rec_name, 
					gont_name = "tl"}))});
	    r2 = ({annotation = "", 
		   body = "return Cons(hd, Nil);\n",
		   args = List::cons1(({yacc_name = of, 
					gont_name = "hd"}))});
    
	    if (!non_empty) {
		_ s2 = ({name = "ne_" + s.name,
			 ty = s.ty,
			 terminal = false,
			 terminal_name = "",
			 rules = List::cons2(r1, r2),
			 id = 0});
		s2.rules = List::cons2(r1, r2);
		Hashtbl::add(symbols, s2.name, s2);
		symbol_list = Cons(s2, symbol_list);
		
		r1 = ({annotation = "", 
		       body = "return Nil;\n",
		       args = Nil});
		
		r2 = ({annotation = "", 
		       body = "",
		       args = List::cons1(({yacc_name = "ne_" + s.name, 
					    gont_name = ""}))});
	    }
	}
	
	s.rules = List::cons2(r1, r2);
    }
    
    void read_optional(symbol s)
    {
	_ of = get_id(l);

	try {
	    let o = Hashtbl::find(symbols, of) in
	    s.ty = "<" + o.ty + ">option";
	} with {
	case Not_found:
	    die("no such symbol " + of);
	}
	
	_ r1, r2;
	
	r1 = ({annotation = "", 
	       body = "return Some(hd);\n",
	       args = List::cons1(({yacc_name = of, 
				    gont_name = "hd"}))});
	r2 = ({annotation = "", 
	       body = "return None;\n",
	       args = Nil});
	
	s.rules = List::cons2(r1, r2);

	switch get_token(l) {
	case Semicolon:
	    skip;
	case _:
	    die("expected ;");
	}
    }
    
    void read_special(symbol s)
    {
	s.terminal = false;

	switch get_id(l) {
	case "list":
	    read_list(s);
	    
	case "optional":
	    read_optional(s);
	
	case sp:
	    die("bad special " + sp);
	}
    }

    _ nam;
    
    try {
	nam = get_id(l);
    } with {
    case End_of_file:
	raise Not_found;
    }

    symbol s = {name = nam, ty = "", terminal = true, 
		rules = Nil, terminal_name = "", id = 0};

    t = get_token(l);
    
    switch t {
    case Body(x):	
	s.ty = x;
	t = get_token(l);
	
    case Semicolon:
    case Name(_):
	// it shouldn't be used
	s.ty = "!bad_type";
    
    case Colon:
	read_special(s);
	// skip next sections
	t = Semicolon;

    case _:		
	die("expected {type}");
    }

    switch t {
    case Name(x):
	s.terminal_name = x;
	t = get_token(l);
	
    case _:
	skip;
    }

    switch t {
    // terminal:
    case Semicolon:
	skip;
    
    // non-terminal
    case Colon:
	while (read_rule(l, s))
	    skip;
	s.rules = List::rev(s.rules);
	s.terminal = false;
    
    case _:
	die("expected ; or :");
    }

    Hashtbl::add(symbols, s.name, s);
    symbol_list = Cons(s, symbol_list);
}

void yacc_symbol(FILE f, symbol s)
{
    int cnt = 0;
    
    void yacc_rule(rule r)
    {
	string hd = "\t  ";

	if (cnt != 0)
	    hd = "\t| ";
	
	let args = List::map(fun _ (rhs_entry e) (e.yacc_name), r.args) in
	fputs(f, hd + String::concat(" ", args) + r.annotation + "\n");

	if (r.body == "") {
	    switch r.args {
	    case Cons(rr, Nil):
		rr.gont_name = "hd";
		r.body = "return hd;\n";
	    case _:
		die("non-1 entry for empty body in " + s.name);
	    }
	}

	int pos = 0;
	
	string maybe_arg(rhs_entry e) 
	{
	    pos++;
	    if (e.gont_name != "")
		return "$" + itoa(pos);
	    else
		return "";
	}

	_ args = List::filter(fun (x) (x != ""), List::map(maybe_arg, r.args));
	_ loc_assign = "";
	
	if (location_type != "") {
	    string loc = "@$.loc_ptr";
	    
	    if (List::length(r.args) == 0)
		loc_assign = "\t\t\t@$ = gyy_last_loc;\n";
		
	    args = Cons(loc, args);
	}

	let call_args = String::concat(", ", args) in 
	let fnname = prefix + s.name + "___" + itoa(cnt++) in
	{
	    _ proto = "extern void *" + fnname + "(" +
		      String::concat(", ", List::map(fun _ (string) ("void*"), 
						     args)) +
		      ");\n";
	    
	    fputs(f, "\t\t{\n\t\t\t" + proto + loc_assign +
		     "\t\t\t$$ = " + fnname + "(" + call_args + ");\n\t\t}\n");
	}
    }
   
    if (s.terminal)
	return;

    fputs(f, s.name + ":\n");
    List::iter(yacc_rule, s.rules);
    fputs(f, ";\n\n");
}

void gont_symbol(FILE f, symbol s)
{
    int cnt = 0;
    
    void gont_rule(rule r)
    {
	string arg(rhs_entry e) 
	{
	    if (e.gont_name != "")
		try {
		    symbol x = Hashtbl::find(symbols, e.yacc_name);
		    return x.ty + " " + e.gont_name;
		} with {
		case Not_found:
		    return die("cannot find rule for " + e.yacc_name);
		}
	    else
		return "";
	}

	_ sargs = List::map(arg, r.args);

	if (location_type != "") {
	    sargs = Cons(location_type + " loc", sargs);
	    r.body = "ignore(loc);\t\n" + r.body;
	}
	    
	let args = 
	    String::concat(", ", List::filter(fun (x) (x != ""), sargs)) in
	let fnname = prefix + s.name + "___" + itoa(cnt++) in
	let header = s.ty + " " + fnname + "(" + args + ") extern \"" 
		     + fnname + "\"" in
	fputs(f, header + "\n{\n\t" + r.body + "}\n\n");
    }
    
    List::iter(gont_rule, s.rules);
}


void main()
{
    if (Array::length(Sys::argv) != 2)
	die("USAGE: " + Array::get(Sys::argv, 0) + " filename\n");
    
    lexer l = make_lexer(fopen(Array::get(Sys::argv, 1), "r"));
   
    for (;;) {
	try { read_symbol(l); } 
	with { case Not_found: break; }
    }

    if (start_symbol == "")
	die("no %start");

    _ syms = List::rev(symbol_list);

    FILE f, f2;

    f = fopen("c_parser.y", "w");
    fputs(f, "/* Autogenerated. */\n"
	     "%{\n"
	     "#include <gont/runtime.h>\n\n"
	     "#define YYSTYPE void*\n"
	     "static int gyy_ret;\n");
    
    if (location_type != "") {
	fputs(f, "#define YYLTYPE gyyltype\n"
		 "typedef struct gyyltype {\n"
		 "	void *loc_ptr;\n"
		 "	int first_line, first_column, last_line, last_column;\n"
		 "} gyyltype;\n"
		 "#define YYLTYPE_IS_TRIVIAL 1\n"
		 "static YYLTYPE gyy_last_loc;\n"
		 "extern YYLTYPE yylloc;\n"
		 "extern void *yylval;\n"
		 "#define GYY_LAST_LOC gyy_last_loc.loc_ptr\n");
    
	fputs(f, "void " + prefix + "_store_lexeme(gont_int ret, "
		 "void *val, gont_int fl, gont_int fc, "
			    "gont_int ll, gont_int lc, void *loc_ptr)\n{\n"
		 "  gyy_last_loc.first_line = fl;\n"
		 "  gyy_last_loc.first_column = fc;\n"
		 "  gyy_last_loc.last_line = ll;\n"
		 "  gyy_last_loc.last_column = lc;\n"
		 "  gyy_last_loc.loc_ptr = loc_ptr;\n"
		 "  yylloc = gyy_last_loc;\n");
    } else {
	fputs(f, "void " + prefix + "_store_lexeme(gont_int ret, "
		 "void *val)\n{\n");
    }

    fputs(f, "	gyy_ret = ret; yylval = val;\n}\n"
	     "extern void " + prefix + "_get_lexeme(void);\n"
	     "static int yylex(void)\n"
	     "{\n"
	     "	" + prefix + "_get_lexeme();\n"
	     "	return gyy_ret;\n"
	     "}\n\n");
    fputs(f, "\nvoid *" + prefix + "_retval;\n");
	     
    fputs(f, c_hd);
    fputs(f, "\n%}\n\n");

    fputs(f, y_hd);
    fputs(f, "\n\n/* Terminals. */\n\n");
    int id = 257;
    void yacc_terminal(symbol s)
    {
	if (s.terminal) {
	    s.id = id;
	    _ tn = "";
	    if (s.terminal_name != "")
		tn = " \"" + s.terminal_name + "\"";
	    fputs(f, "%token " + s.name + " " + itoa(id++) + tn + "\n");
	}
    }

    List::iter(yacc_terminal, syms);
    fputs(f, "\n/* End of terminals. */\n\n%start gyacc_start\n\n");
    fputs(f, y_tl);
    fputs(f, "\n\n%%\n");
    fputs(f, "gyacc_start: " + start_symbol + 
	     "\n\t{ " + prefix + "_retval = $1; };\n\n");

    
    List::iter(fun (s) (yacc_symbol(f, s)), syms);
    fputs(f, "\n%%\n\n");
    fputs(f, "void *" + prefix + "_parse() {\n"
	     "\treturn (void*)((long)yyparse());\n"
	     "}\n\n");
    fputs(f, c_tl);
    fputs(f, "/* EOF */\n");
    fclose(f);

    f = fopen("parser.g", "w");
    fputs(f, "/* Autogenerated. */\n");
    fputs(f, g_hd);
    List::iter(fun (s) (gont_symbol(f, s)), syms);
    f2 = fopen("parser.gi", "w");
    fputs(f2, "/* Autogenerated. */\n");
    fputs(f2, gi_hd);
    try {
	symbol s = Hashtbl::find(symbols, start_symbol);
	_ lex_fnt;
	if (location_type == "")
	    lex_fnt = "*(" + module + "::token ())";
	else
	    lex_fnt = "*(*(" + module + "::token, location) ())";
	fputs(f, "extern int c__parse() = \"" + prefix + "_parse\";\n"
		 "extern " + s.ty + " c__retval = \"" + prefix + "_retval\";\n"
		 + s.ty + " parse(" + lex_fnt + " l)\n"
		 "{\n"
		 "  gy_lex_fn = l;\n"
	         "  c__parse();\n"
		 "  if (Hacks::is_null(c__retval))\n"
		 "	raise Failure(\"" + module + "::parse\");\n"
		 "  return c__retval;\n"
		 "}\n"
		 );
	fputs(f, "def gy_choke() {return Sys::abort();}\n"
		 + lex_fnt + " gy_lex_fn = gy_choke;\n"
		 "extern void c__store_lexeme(int, int, int, int, int, int, int)"
				    " = \"" + prefix + "_store_lexeme\";\n"
		 "void get_lexeme() extern \"" + prefix + "_get_lexeme\"\n"
		 "{\n"
		 "\t_ fl, fc, ll, lc, loc_ptr;\n"
		 "\t_ ret, val = 0, tok;\n");
	if (location_type == "") {
	    fputs(f, "\tfl = fc = ll = lc = loc_ptr = 0;\n"
		     "\ttok = gy_lex_fn();\n");
	} else {
	    fputs(f, "\t_ loc;\n"
		     "\t(tok, loc) = gy_lex_fn();\n"
		     "\tfl = loc.first_line;\n"
		     "\tfc = loc.first_column;\n"
		     "\tll = loc.last_line;\n"
		     "\tlc = loc.last_column;\n"
		     "\tloc_ptr = Hacks::ptr_of(loc);\n");
	}
	fputs(f, "\tswitch tok {\n"
		 "\tcase " + module + "::Keyword(n):\n"
		 "\tcase " + module + "::Operator(n):\n"
		 "\t\tret = n;\n"
		 "\tcase " + module + "::Eof:\n"
		 "\t\tret = 0;\n");
			
	// cannot have Parser:: here
	if (location_type == "")
	    lex_fnt = "*(token ())";
	else
	    lex_fnt = "*(*(token, location) ())";
	fputs(f2, "\n\n" +
		  s.ty + " parse(" + lex_fnt + " lex_fn);\n\n"
		  "<*(string, int)>list operators, keywords;\n\n"
	          "union token {\n"
		  "\tint Keyword;\n"
		  "\tint Operator;\n"
		  "\tvoid Eof;\n");
		  
	def parametrized_terminal(_, s)
	{
	    if (s.terminal && s.ty != "!bad_type") {
		fputs(f2, "\t" + s.ty + " " + s.name + ";\n");
		fputs(f, "\tcase " + module + "::" + s.name + "(v):\n"
			 "\t\tret = " + itoa(s.id) + ";\n"
			 "\t\tval = Hacks::ptr_of(v);\n");
	    }
	}

	Hashtbl::iter(symbols, parametrized_terminal);
		
	fputs(f2, "}\n");

	fputs(f, "\t}\n"
		 "\tc__store_lexeme(ret, val, fl, fc, ll, lc, loc_ptr);\n"
		 "}\n");
		 
	fputs(f, "<*(string, int)>list operators = [\n");
	bool is_first = true;
	
	def yacc_c1_terminal(c, _)
	{
	    _ i = int_of_char(c);
	    if (i > 0xff)
		die("non-ascii one-char terminal");
	    if (!is_first) fputs(f, ",");
	    is_first = false;
	    fputs(f, "(\"" + String::make(1, c) + "\", " + itoa(i) + ")\n");
	}

	def push_one(op, s)
	{
	    _ n = s.terminal_name;
	    if (n != "") {
		_ is_op = !(Unitype::is_alnum(String::get(n, 0)) || n == "_");
		if (op && !is_op) return;
		if (!op && is_op) return;
		if (!is_first) fputs(f, ",");
		is_first = false;
		fputs(f, "(\"" + n + "\", " + itoa(s.id) + ")\n");
	    }
	}
	
	if (Ptrhash::mem(one_char_terminals, '_')) {
	    Ptrhash::remove(one_char_terminals, '_');
	    Hashtbl::add(symbols, "blah", 
			 {terminal_name = "_", id = int_of_char('_'),
			  ty = "!bad_type", name = "ble", rules = [],
			  terminal = true});
	}

	is_first = true;
	Hashtbl::iter(symbols, fun (_, s) (push_one(true, s)));
	Ptrhash::iter(one_char_terminals, yacc_c1_terminal);
	fputs(f, "];\n<*(string, int)>list keywords = [\n");
	is_first = true;
	Hashtbl::iter(symbols, fun (_, s) (push_one(false, s)));
	fputs(f, "];\n");
	
    } with {
    case Not_found:
	die("cannot locate start symbol " + start_symbol);
    }
    fputs(f, g_tl);
    fputs(f, "\n/* EOF */\n");
    fclose(f);
    fputs(f2, gi_tl);
    fputs(f2, "\n/* EOF */\n");
    fclose(f2);
}

'a die(string msg)
{
    puts_err("fatal error: " + msg + "\n"
	     "line: " + itoa(lines_read) + "\n"
	     "bailing out.\n");
    return Sys::exit(1);
}

section init
{
    try {
	main();
    } with {
    case Sys_error(m):
	die("sys_error " + m);
	
    case End_of_file:
	die("EOF");
	
    case Not_found:
	die("not found");

    case _:
	die("unknown exception");
    }
}

// vim:sts=4:sw=4
